#include <data_priv.h>
#include <data_type.h>
#include <data_set.h>

#include <stdlib.h>
#include <string.h>

static int
internalstore(data_core_t core)
{
	int t;

	if (core->dc_type->dt_size > (roundpow2(core->dc_set->ds_elementsize) -
	    core->dc_set->ds_elementsize + sizeof (void *)))
		return 0;
	return 1;
}

void
d_copyin(data_core_t core, void *a, void *b)
{
	void *ptr;
	(char *)b += core->dc_set->ds_elementsize - sizeof (void *);
	if (internalstore(core))
		memmove(b, a, core->dc_type->dt_size);
	else {
		ptr = malloc(core->dc_type->dt_size);
		memmove(ptr, a, core->dc_type->dt_size);
		memmove(b, &ptr, sizeof(void *));
	}
}

void
d_copydest(data_core_t core, void *a)
{
	(char *)a += core->dc_set->ds_elementsize - sizeof (void *);
	if (!internalstore(core))
		free(*(void **)a);
	*(void **)a = NULL;
}

void *
d_copyout(data_core_t core, void *a, int event, void *co_arg)
{
	int size;
	void *ret;
	void *ptr;

	size = core->dc_type->dt_size;
	ptr = NULL;

	if (a == NULL) {
		if (core->dc_flags & D_PARAMPTR)
			*(void **)co_arg = NULL;
		return NULL;
	}

	(char *)a += core->dc_set->ds_elementsize - sizeof (void *);
	if (core->dc_flags & D_RETVAL) {
		ptr = &ret;
		size = sizeof ret;
	} else if (core->dc_flags & D_PARAMPTR)
		ptr = co_arg;
	else
		ptr = malloc(size);

	if (internalstore(core))
		memmove(ptr, a, size);
	else
		memmove(ptr, *(void **)a, size);

	if (core->dc_flags & D_RETVAL)
		return ret;

	return ptr;
}

void *
d_dptr(data_core_t core, void *a)
{
	(char *)a += core->dc_set->ds_elementsize - sizeof (void *);
	if (internalstore(core))
		return a;
	return *(void **)a;
}
