#include <config.h>
#include <data_set.h>
#include <data_priv.h>

#include <dlfcn.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static data_set_t
searchfile(char *file, char *sname, data_core_t core)
{
	void *dlptr;
	data_set_t set;
	char buf[256];
	char name[256];
	char lib[256];
	char *libname;
	FILE *fp;
	int i;

	if ((fp = fopen(file, "r")) == NULL)
		return NULL;

	set = NULL;
	while (set == NULL && !feof(fp) && !ferror(fp)) {
		if (fgets(buf, 256, fp) == NULL || buf[0] == '#')
			continue;
		i = strlen(buf) - 1;
		if (buf[i] == '\n')
			buf[i] = '\x0';
		if (sscanf(buf, "%s %s", name, lib) != 2 || strcmp(name, sname))
			continue;

		if ((dlptr = dlopen(lib, RTLD_NOW)) == NULL) {
			printf("open of lib %s failed: %s\n", lib, dlerror());
			continue;
		}
		core->dc_libptr = dlptr;
		/* grab symbol from library, make sure it matches */
		if ((set = (data_set_t) dlsym(dlptr, "_d_set")) == NULL ||
		    strcmp(sname, set->ds_name)) {
			dlclose(dlptr);
			set = NULL;
		}
	}
	return set;
}

static data_set_t
getname(char *sname, data_core_t core)
{
	data_set_t set;

	set = searchfile(_PATH_DATA_CONF, sname, core);
	/* when we have more places to search, add 'em hear */

	return set;
}

data_core_t
d_create(char *sname, data_type_t type, int flags)
{
	data_core_t core;
	data_set_t set;

	if ((core = (data_core_t) malloc(sizeof *core)) == NULL)
		return NULL;
	if ((set = getname(sname, core)) == NULL)
		return NULL;

	core->dc_data = NULL;
	core->dc_set = set;
	core->dc_type = type;
	core->dc_flags = flags;
	core->dc_set->ds_handler(core, D_INIT, NULL, NULL);

	return core;
}

void *
d_sendevent(data_core_t core, int event, ...)
{
	void *co_arg;
	void *ret;
	va_list ap;

	va_start(ap, event);
	co_arg = NULL;

	switch (event) {
	case D_INIT:
		/*
		 * these are handled internally and shouldn't be
		 * sent by the user's code.
		 */
		ret = NULL;
		break;
	default:
		/*
		 *we might want to make sure that a set actually
		 * supports the event we are sending it.
		 */
		/*
		 * we also need to define a standard on how multiple
		 * values are returned.  Probably just a void **
		 * that contains each void *
		 */
		if (core->dc_flags & D_PARAMPTR)
			co_arg = va_arg(ap, void *);
		ret = core->dc_set->ds_handler(core, event, co_arg, ap);
		break;
	}
	va_end(ap);
	return ret;
}

static int
checkflags(int flags)
{
	if (flags & (D_RETVAL | D_PARAMPTR)) {
		errno = EINVAL;
		return -1;
	}
	return 0;
}

int
d_setflags(data_core_t core, int flags)
{
	flags = core->dc_flags | flags;
	if (checkflags(flags) == -1)
		return -1;
	core->dc_flags = flags;
	return 0;
}

int
d_clearflags(data_core_t core, int flags)
{
	flags = core->dc_flags & ~flags;
	if (checkflags(flags) == -1)
		return -1;
	core->dc_flags = flags;
	return 0;
}
