#include <data_priv.h>
#include <data_set.h>
#include "fib.h"

#include <stdarg.h>
#include <stdlib.h>

static void *ds_fib_handler(data_core_t core, int event, void *co_arg,
    va_list ap);

struct data_set d_set = {
	"fib",
	sizeof(struct fib_heap),
	sizeof(struct fib_heap_el),
	ds_fib_handler
};

static void *
ds_fib_handler(data_core_t core, int event, void *co_arg, va_list ap)
{
	struct fib_heap *h;
	struct fib_heap_el *nptr;
	struct fib_heap_el *x, *y, *orig;
	void *ptr;

	h = core->dc_data;
	switch (event) {
	case D_INIT:
		if ((h = core->dc_data = fh_makeheap()) == NULL) {
			printf("heap is null\n");
			exit(1);
		}
		fh_setcmp(h, core->dc_type->dt_comp);
		fh_setneginf(h, core->dc_type->dt_neginf);
		break;
	case D_ADDKEY:
		nptr = malloc(roundpow2(sizeof(struct fib_heap_el)));
		fhe_initelem(nptr);
		d_copyin(core, va_ptrarg(ap, char[core->dc_type->dt_size]),
		    nptr);
		printf("heap: %p, num: %d, min ptr: %p, root: %p\n", h, h->fh_n,
		    h->fh_min, h->fh_root);
		fh_insertrootlist(h, nptr);

		if (h->fh_min == NULL || h->fh_cmp_fnct(d_dptr(core, nptr),
		    d_dptr(core, h->fh_min)) < 0)
			h->fh_min = nptr;
		h->fh_n++;
		break;
	case D_MIN:
		nptr = fh_min(h);
		return d_copyout(core, nptr, event, co_arg);
		break;
	case D_DELETEMIN:
		/*
		 * this is effectivly a duplicate of the code in
		 * fh_extractmin
		 */
		printf("deletemin\n");
		nptr = h->fh_min;
		ptr = d_copyout(core, nptr, event, co_arg);
		if (nptr == NULL)
			return ptr;
		orig = NULL;
		printf("dumping children\n");
		for(x = nptr->fhe_child; x != orig && x != NULL;) {
			if (orig == NULL)
				orig = x;
			y = x->fhe_right;
			x->fhe_p = NULL;
			fh_insertrootlist(h, x);
			x = y;
		}
		printf("removing node\n");
		fh_removerootlist(h, nptr);
		h->fh_n--;
		printf("updating min ptr\n");
		if (nptr == nptr->fhe_right)
			h->fh_min = NULL;
		else {
			h->fh_min = nptr->fhe_right;
			fh_consolidate(h);
		}
		d_copydest(core, nptr);
		free(nptr);
		return ptr;
		break;
	}
	return NULL;
}
