#include <data_priv.h>
#include <data_set.h>

#include <sys/queue.h>
#include <stdarg.h>
#include <stdlib.h>

typedef void (*operate_on_data)(void *);

SLIST_HEAD(slist_head, slist_node);

struct slist_node {
	SLIST_ENTRY(slist_node) nd;
	void *data;
};

static void *ds_slist_handler(data_core_t core, int event, void *co_arg,
    va_list ap);

struct data_set d_set = {
	"slist",
	sizeof(struct slist_head),
	sizeof(struct slist_node),
	ds_slist_handler
};

static void *
ds_slist_handler(data_core_t core, int event, void *co_arg, va_list ap)
{
	struct slist_node *nptr, *pptr, *iptr;
	void *ptr;
	operate_on_data f;

	nptr = pptr = iptr = NULL;
	ptr = NULL;

	switch (event) {
	case D_INIT:
		SLIST_INIT((struct slist_head *)&(core->dc_data));
		break;
	case D_INSERT_HEAD:
		nptr = malloc(roundpow2(sizeof(struct slist_node)));
		d_copyin(core, va_ptrarg(ap, char[core->dc_type->dt_size]),
		    nptr);
		SLIST_INSERT_HEAD((struct slist_head *)&core->dc_data, nptr,
		     nd);
		break;
	case D_REMOVE_HEAD:
		if ((nptr =
		    SLIST_FIRST((struct slist_head *)&core->dc_data)) == NULL)
			break;
		d_copydest(core, nptr);
		SLIST_REMOVE_HEAD((struct slist_head *)&core->dc_data, nd);
		free(nptr);
		break;
	case D_FIRST:
		nptr = SLIST_FIRST((struct slist_head *)&core->dc_data);
		return d_copyout(core, nptr, event, co_arg);
		break;
	case D_INSERT_AFTER:
		/* we will insert the second one after the first one */
		ptr = va_ptrarg(ap, char[core->dc_type->dt_size]);
		nptr = SLIST_FIRST((struct slist_head *)&core->dc_data);
		while (nptr != NULL && core->dc_type->dt_comp(ptr, d_dptr(core,
		    &nptr->data))) {
			pptr = nptr;
			nptr = SLIST_NEXT(nptr, nd);
		}
		iptr = malloc(roundpow2(sizeof(struct slist_node)));
		d_copyin(core, va_ptrarg(ap, char[core->dc_type->dt_size]),
		    iptr);
		if (nptr == NULL) {
			SLIST_INSERT_AFTER(pptr, iptr, nd);
		} else {
			SLIST_INSERT_AFTER(nptr, iptr, nd);
		}
		break;
	case D_REMOVE:
		nptr = SLIST_FIRST((struct slist_head *)&core->dc_data);
		while (nptr != NULL && core->dc_type->dt_comp(ptr, d_dptr(core,
		    &nptr->data))) {
			pptr = nptr;
			nptr = SLIST_NEXT(nptr, nd);
		}
		if (nptr == NULL)
			break;	/* data element wasn't on list!! */
		pptr->nd.sle_next = nptr->nd.sle_next;
		d_copydest(core, nptr);
		free(nptr);
		break;
	case D_HEADPTR:
		return &core->dc_data;
		break;
	case D_NEXTPTR:
		nptr = va_ptrarg(ap, char[sizeof(void *)]);
		return SLIST_NEXT(nptr, nd);
		break;
	case D_PTRDATA:
		nptr = va_ptrarg(ap, char[sizeof(void *)]);
		return d_copyout(core, nptr, event, co_arg);
		break;
	case D_FOREACH:
		nptr = SLIST_FIRST((struct slist_head *)&core->dc_data);
		f = va_ptrarg(ap, operate_on_data);
		while (nptr != NULL)
			f(d_dptr(core, &nptr->data));
		break;
	}
	return NULL;
}
