#include <data_priv.h>
#include <data_set.h>

#include <sys/queue.h>
#include <stdarg.h>
#include <stdlib.h>

SLIST_HEAD(stack_head, stack_node);

struct stack_node {
	SLIST_ENTRY(stack_node) nd;
	void *data;
};

static void *ds_stack_handler(data_core_t core, int event, void *co_arg,
    va_list ap);

static void *
ds_stack_handler(data_core_t core, int event, void *co_arg, va_list ap)
{
	struct stack_node *nptr;
	void *ptr;

	switch (event) {
	case D_INIT:
		SLIST_INIT((struct stack_head *)&(core->dc_data));
		break;
	case D_PUSH:
		nptr = malloc(roundpow2(sizeof(struct stack_node)));
		d_copyin(core, va_ptrarg(ap, char[core->dc_type->dt_size]),
		    nptr);
		SLIST_INSERT_HEAD((struct stack_head *)&core->dc_data, nptr,
		     nd);
		break;
	case D_FIRST:
		nptr = SLIST_FIRST((struct stack_head *)&core->dc_data);
		return d_copyout(core, nptr, event, co_arg);
		break;
	case D_POP:
		nptr = SLIST_FIRST((struct stack_head *)&core->dc_data);
		ptr = d_copyout(core, nptr, event, co_arg);
		if (nptr != NULL) {
			d_copydest(core, nptr);
			SLIST_REMOVE_HEAD((struct stack_head *)&core->dc_data,
			    nd);
			free(nptr);
		}
		return ptr;
		break;
	}
	return NULL;
}

struct data_set d_set = {
	"stack",
	sizeof(struct stack_head),
	sizeof(struct stack_node),
	ds_stack_handler
};
